/* This program compresses a file without losing information.
 * The unsqueeze program is required to unsqueeze the file
 * before it can be used.
 *
 * If the name of a file to be squeezed has a CP/M, DOS, HDOS, usw.
 * .XXX extension, the second letter is changed to "q", otherwise
 * ".qqq" is appended.  If the name exists it is overwritten!
 *
 * The transformations compress strings of identical bytes and
 * then encode each resulting byte value as bit strings
 * having lengths in inverse proportion to their frequency of
 * occurrance in the intermediate input stream. The latter uses
 * the Huffman algorithm.  Decoding information is included in
 * the squeezed file, so squeezing short files or files with
 * uniformly distributed byte values will actually increase size.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "sqcom.h"
#include "sq.h"
#include "sq_help.h"

char outfile[2048],	/* output file spec. */
     outdir[1024];

int
main(int argc, char **argv)
{
	int i,c;
	char inparg[16];	/* parameter from input */

	debug = FALSE;
	outdir[0] = '\0';

	while ((c = getopt(argc,argv,"Vvho:")) != EOF) {
	  switch(c) {
	    case 'V':
	      display_version();
	      return 0;

	    case 'v':
	      debug = TRUE;
	      break;

            case 'o':
	      strncpy(outdir,optarg,1022);
	      outdir[1022] = '\0';  /* make sure string is terminated */
	      i = strlen(outdir);
	      if (i)
	        if (outdir[i-1] != '/') strcat(outdir,"/");
	      break;

	    case 'h':
	    default:
	      display_help();
	      return 0;
	  } /* switch */
	} /* while */
	      
	if (optind >= argc) {
		printf("File squeezer " VERSION " \n"
		       "Conceived by Richard Greenlaw Modified by Chuck Forsberg et al.\n"
		       "Parameters (from command line or singly from stdin)\n"
		       "consist of output drives and input file names.\n");
		do {
			printf("\n*");
			for(i = 0; i < 16; ++i) {
				if((c = getchar()) == EOF)
					c = '\n';	/* fake empty (exit) command */
				if((inparg[i] = c) == '\n') {
					inparg[i] = '\0';
					break;
				}
			}
			if(inparg[0] != '\0')
				obey(inparg);
		} while(inparg[0] != '\0');
	} else {
		for (i = optind; i < argc; i++) obey(argv[i]);
	}
	return 0;
}

void obey(char outname[])
{
	int l,i;

	/* build output file name */
	outfile[0] = '\0';		/* empty */
        strcat(outfile, outdir);
	strcat(outfile, outname);

	/* check for the presence of a 1-3 char "extension" */
	for (l = i = strlen(outfile) - 1; (i >= 1) && (outfile[i] != '.'); i--)
	  ;
	if (i) {
	  switch (l - i) {
	    case 0:
	      strcat(outfile,"qqq");
	      break;
	    case 1:
	      strcat(outfile,"q");
	      break;
	    case 2:
	      if (outfile[l] == 'q')
	        strcat(outfile,".qqq");
	       else
	        outfile[l] = 'q';
	      break;
	    case 3:
	      if (outfile[l-1] == 'q')
	        strcat(outfile,".qqq");
	       else
	       outfile[l-1] = 'q';
	      break;
	    default:
	      strcat(outfile,".qqq");
	      break;
	  } /* switch */
	} else {
	  strcat(outfile,".qqq");
	}

	squeeze(outname, outfile);
}

void
squeeze(char *infile, char *outfile)
{
	int c;

	printf("%s -> %s: ", infile, outfile);

	if(!(ifp = fopen(infile, "r"))) {
		fprintf(stderr, "Can't open %s\n", infile);
		return;
	}
	if(!(ofp = fopen(outfile, "w"))) {
		fprintf(stderr, "Can't create %s\n", outfile);
		fclose(ifp);
		return;
	}

	/* First pass - get properties of file */
	crc = 0;	/* initialize checksum */
	init_ncr();
	printf("scanning, ");
	fflush(stdout);
	init_huff(ifp);   
	fclose(ifp);

	/* Write output file header with decoding info */
	wrt_head(ofp,infile);

	/* Second pass - encode the file */
	printf("squeezing, ");
	fflush(stdout);
	if(!(ifp = fopen(infile, "r"))) {
		fprintf(stderr, "Can't open %s\n", infile);
		goto closeout;
	}
	init_ncr();	/* For second pass */

	/* Translate the input file into the output file */
	while((c = gethuff(ifp)) != EOF)
		if(putc(c, ofp) == EOF) {
			fprintf(stderr, "ERROR - write error in %s\n", outfile);
			goto closeall;
		}
	printf(" done.\n");
closeall:
	fclose(ifp);
closeout:
	fclose(ofp);
}
