/* USQ.C CHANGE HISTORY:
 *
 * 0.1  Ported from original CP/M usq v2.0
 *
 * Program to unsqueeze files formed by CP/M sq.com, usw.
 */
#include "config.h"
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "sqcom.h"
#include "usq.h"
#include "get_put.h"
#include "usq_help.h"
#include "dnode_entry.h"

int	dispcnt;	/* How much of each file to preview */
int	ffflag;		/* true if ff separates preview from different files */
int	nlmode;		/* true if adding/deleting cr's for host system */
int	lowercase;	/* true if casting output filename to lowercase */
int	crlfmode;	/* true for crlf->lf conversion */
int	to_stdout;	/* true if output goes to stdout */
char	outpath[1024];	/* output path */

static void unsqueeze(char *infile)
{
	FILE *inf, *outf;	/* file pointers */
	int i, c, from_stdin = (infile == NULL);
	char cc;
	int suppress = 0;

	char *p;
	unsigned filecrc;	/* checksum */
	int numnodes;		/* size of decoding tree */
	char outfile[2048];	/* output file name */
	unsigned linect;	/* count of number of lines previewed */

	if (from_stdin) {
	  inf = stdin;
	} else {
	  if (!(inf = fopen(infile,"r"))) {
		  fprintf(stderr, "Can't open %s\n", infile);
		  return;
	  }
	}

	/* Initialization */
	linect = 0;
	crc = 0;
	init_cr();
	init_huff();

	/* Process header */
	if(q_get_int16(inf) != RECOGNIZE) {
		if (!from_stdin)
		  fprintf(stderr,"%s is ",infile);
		fprintf(stderr, "not a squeezed file\n");
		goto closein;
	}

	filecrc = q_get_int16(inf);

	/* Get original file name */
	p = origname;	/* send it to array */
	do {
		*p = lowercase?tolower(getc(inf)):getc(inf);
	} while(*p++ != '\0');

	/* Combine with output path */
	outfile[0] = '\0';		/* empty */
	strcat(outfile,outpath);	/* path */
	strcat(outfile, origname);	/* name */

	if (!to_stdout) {
	  if (!from_stdin) printf("%s ",infile);
	  printf("-> %s: ", outfile);
	  fflush(stdout);
	}

	numnodes = q_get_int16(inf);

	if(numnodes < 0 || numnodes >= NUMVALS) {
		fprintf(stderr, "%s has invalid decode tree size\n", infile);
		goto closein;
	}

	/* Initialize for possible empty tree (SPEOF only) */
        *dnode_entry(0,0) = -(SPEOF + 1);
	*dnode_entry(0,1) = -(SPEOF + 1);

/*
	dnode[0].children[0] = -(SPEOF + 1);
	dnode[0].children[1] = -(SPEOF + 1);
*/
	/* Get decoding tree from file */
	for(i = 0; i < numnodes; ++i) {
		*dnode_entry(i,0) = q_get_int16(inf);
		*dnode_entry(i,1) = q_get_int16(inf);
/*
		dnode[i].children[0] = q_get_int16(inf);
		dnode[i].children[1] = q_get_int16(inf);
*/
	}

	if(dispcnt) {
		/* Use standard output for previewing */
		putchar('\n');
		while(((c = getcr(inf)) != EOF) && (linect < dispcnt)) {
			cc = 0x7f & c;	/* strip parity */
			if((cc < ' ') || (cc > '~'))
				/* Unprintable */
				switch(cc) {
				case '\r':	/* return */
					/* newline will generate CR-LF */
					goto next;
				case '\n':	/* newline */
					++linect;
				case '\f':	/* formfeed */
				case '\t':	/* tab */
					break;
				default:
					cc = '.';
				}
			putchar(cc);
		next: ;
		}
		if(ffflag)
			putchar('\f');	/* formfeed */
	} else {
		/* Create output file */
		if (to_stdout) {
		  outf = stdout;
		} else {
		  if(!(outf = fopen(outfile, "w"))) {
			  fprintf(stderr, "Can't create %s\n", outfile);
			  goto closeall;
		  }
		}
		/* Get translated output bytes and write file */
		while((c = getcr(inf)) != EOF) {
			crc += c;
#if (SIZEOF_UNSIGNED > 2)
			if (crc > 0xffff) crc -= (unsigned)0x10000;
#endif

			if(nlmode && c=='\n')
				putc('\r', outf);

			if (crlfmode) {
			  if (suppress)	{	/* prev char a \r? */
			    suppress = 0;
			    if (c != '\n')	/* not crlf, write the cr */
			      putc('\r',outf);
			  }

			  suppress = (c == '\r');
			}

			if (!suppress)
			if((putc(c, outf)) == EOF) {
				fprintf(stderr, "Write error in %s\n", outfile);
				goto closeall;
			}
		}
		if (crlfmode && suppress)
		  putc('\r',outf);	/* write the dangling cr */

		if(filecrc != crc) {
			fprintf(stderr, "ERROR - checksum error in %s\n", outfile);
		}


	closeall:
		if (!to_stdout) fclose(outf);
	}

closein:
	if (!from_stdin) fclose(inf);
	if (!to_stdout) fputc('\n',stdout);
}

int main(int argc, char **argv)
{
	int i,c;

	crlfmode = nlmode = lowercase = dispcnt = 0;
	outpath[0] = '\0';

	while ((c = getopt(argc,argv,"NnLf:p:o:OVh")) != EOF) {
	  switch(c) {
	    case 'N':
	      nlmode = 1;
	      crlfmode = 0;
	      break;

	    case 'n':
	      nlmode = 0;
	      crlfmode = 1;
	      break;

	    case 'L':
	      lowercase = 1;
	      break;

	    case 'f':
	      ffflag = 1;

	    case 'p':
	      dispcnt = atoi(optarg);
	      break;

	    case 'o':
	      strncpy(outpath,optarg,1022);
	      outpath[1022] = '\0';  /* make sure string is terminated */
	      i = strlen(outpath);
	      if (i)
	        if (outpath[i-1] != '/') strcat(outpath,"/");
	      break;

	    case 'O':
	      to_stdout = 1;
	      break;

	    case 'V':
	      display_version();
	      exit(0);

	    case 'h':
	    default:
	      display_help();
	      exit(0);
	  } /*switch*/
	}

	/* hopefully, any remaining args are filenames */

	if (optind >= argc) {
	  unsqueeze(NULL);
	} else {
	  for (i = optind; i < argc; i++)
	    unsqueeze(argv[i]);
	}
	return 0;
}
